from netgen.meshing import *
from ngsolve import *

SB = 20.0       # specimen width
CA = 0.5*SB     # crack length
SH = 0.5*SB     # specimen half-height

mesh=Mesh("SEC_2D_lin_tri.vol")

# Define element order
ele_order = 2

# Perform nref uniform refinements
# nref = 4
# for inum in range(nref):
 # mesh.Refine()

##   Elasticity parameters
E = 207.0e3 # Elastic modulus
nu = 0.3    # Poisson's ratio

## Displacement Loading on top surface
ud_top = 1.0E-02

##   2D Approximation
model = "plane strain"
#model = "plane_stress"

## Alternate Elasticity parameters (dependent on E, nu and 2D approximation)
mu = E/2/(1+nu)
lmbda = E*nu/(1+nu)/(1-2*nu)
if model == "plane_stress":
    lmbda = 2*mu*lmbda/(lmbda+2*mu)

# Define space of functions
#fes1 = H1(mesh=mesh, order=1)
fes1 = FESpace(type="nodal", mesh=mesh, order=ele_order)
fes = FESpace([fes1, fes1])
#print(fes.ndof)
#fes = VectorH1(mesh=mesh, order=ele_order, type="nodal")  # Note: Vector H1 space has no. of components = mesh.dim

# Dirichlet boundary conditions
fesx, fesy = fes.components
freedofs = fes.FreeDofs()
tol = 1.0E-14

pres_load_dofs = [] # Set up array to store DOFs with non-zero Dirichlet BCs

# Apply Dirichlet BCs (uy=0) on boundary marked by "Ysym"
for el in fesx.Elements(BND):
    if el.mat == "Ysym":
        for dof in el.dofs:
            #print(dof)
            freedofs.Clear(fesx.ndof+dof)
    elif el.mat == "Pres_load":
         for dof in el.dofs:
             #print(dof)
             freedofs.Clear(fesx.ndof+dof)
             pres_load_dofs.append(fesx.ndof+dof)


# Mark DOFs for displacement loading as Dirichlet DOFs and store them for later application of non-homgeneous Dirichlet condition "uy=ud_top" on surface "Pres_load".
# for edge in mesh.edges:
    # if all(mesh[v].point[1] > (SH-tol) for v in edge.vertices):
        # # Vertex DOFs
        # for vertex in edge.vertices:
            # for dof in fesx.GetDofNrs(vertex):
                # freedofs.Clear(fesx.ndof+dof)
                # pres_load_dofs.append(fesx.ndof+dof)
        # # Edge DOFs        
        # for dof in fesx.GetDofNrs(edge):
            # freedofs.Clear(fesx.ndof+dof)
            # pres_load_dofs.append(fesx.ndof+dof)



for v in mesh.vertices:
    #print(v, v.point[0], v.point[1], fes.GetDofNrs(v))
    #Apply symmetry BCs (uy=0)  on "Ysym" - part of bottom edge
#    if v.point[0] > (0.0-tol) and v.point[1] < tol:
#        i = 0
#        for dof in fes.GetDofNrs(v):
#           i+=1
#           if i==2: 
#               freedofs.Clear(dof)
#               print(v.point[0], v.point[1], fes.GetDofNrs(v))
           
     
    #Apply ux=0 at the bottom-right corner node
    if v.point[0] > (0.5*SB-tol) and v.point[1] < tol:
         for dof in fesx.GetDofNrs(v):
             freedofs.Clear(dof)
#             print(v.point[0], v.point[1], fesx.GetDofNrs(v))

#print(fes.FreeDofs())

ux, uy = fes.TrialFunction()
vx, vy = fes.TestFunction()

#force = CoefficientFunction( (0.0,100.0) )
#ud = CoefficientFunction(1.0E-04)

#ds = dx(definedon=mesh.Boundaries("Pres_load"))

# some utils ...
def IdentityCF(dim):
    return CoefficientFunction( tuple( [1 if i==j else 0 for i in range(dim) for j in range(dim)]), dims=(dim,dim) )

def Trace(mat):
    return sum( [mat[i,i] for i in range(mat.dims[0]) ])

# Define Strain Tensor
def eps(ux, uy):
    eps11 = grad(ux)[0]
    eps22 = grad(uy)[1]
    eps12 = 0.5*(grad(ux)[1]+grad(uy)[0])
    eps21 = eps12
    return CoefficientFunction((eps11,eps12,eps21,eps22),dims=(2,2))
    #return 0.5*(u.Deriv() + u.Deriv().trans)  # NOTE: row .. component, col .. derivative

# Define Stress Tensor
def sigma(ux, uy):
    return lmbda*Trace(eps(ux, uy))*IdentityCF(2) + 2.0*mu*eps(ux, uy)

# the solution field 
gfu = GridFunction(fes)
gfu.vec[:] = 0.0
for dof in pres_load_dofs:
	gfu.vec[dof] = ud_top

f = LinearForm(fes)
#f += SymbolicLFI(force*v, definedon=mesh.Boundaries("Pres_load"))
#f += SymbolicLFI(InnerProduct(force,v), definedon=mesh.Boundaries("Pres_load"))
#f += InnerProduct(force, v)*ds

# the bilinear-form 
a = BilinearForm(fes, symmetric=True)
#a += SymbolicBFI(InnerProduct(sigma(u), eps(v)))
a += InnerProduct(sigma(ux, uy), eps(vx, vy))*dx

a.Assemble()
f.Assemble()

res = gfu.vec.CreateVector()
#res.data = f.vec - a.mat*gfu.vec
res.data = -a.mat*gfu.vec

# Obtain solution
gfu.vec.data += a.mat.Inverse(fes.FreeDofs(), inverse="umfpack") * res

fes_sig = L2(mesh, order=1, dim=2*mesh.dim)
gf_sig =  GridFunction(fes_sig,"stress")

gfux, gfuy = gfu.components
gf_sig.Set(sigma(gfux, gfuy))

for dof in pres_load_dofs:
	print(dof, gfu.vec[dof])

print("Y-displacement at vertex DOFs on top surface")
print("uy at x,y,= -10,10.0: ", gfuy(mesh(-10.0,10.0)))
print("uy at x,y,= -5,10.0: ", gfuy(mesh(-5.0,10.0)))
print("uy at x,y,= 0,10.0: ", gfuy(mesh(0.0,10.0)))
print("uy at x,y,= 5,10.0: ", gfuy(mesh(5.0,10.0)))
print("uy at x,y,= 10,10.0: ", gfuy(mesh(10.0,10.0)))

print("uy at x,y,= -7.5,10.0: ", gfuy(mesh(-7.5,10.0)))
print("uy at x,y,= -2.5,10.0: ", gfuy(mesh(-2.5,10.0)))
print("uy at x,y,= 2.5,10.0: ", gfuy(mesh(2.5,10.0)))
print("uy at x,y,= 7.5,10.0: ", gfuy(mesh(7.5,10.0)))

Draw (gfux, mesh, "dispx")
Draw (gfuy, mesh, "dispy")
Draw (gf_sig, mesh, "stress")
SetVisualization (deformation=True)

# VTKOutput object
vtk = VTKOutput(ma=mesh,
                coefs=[gfux, gfuy, gf_sig],
                names = ["dispx", "dispy", "stress"],
                filename="SEC_2D_tri_result",
                subdivision=0)
# Exporting the results:
vtk.Do()